package org.greenstone.gsdl3.util;

import java.util.ResourceBundle;
import java.util.Locale;
import java.util.Enumeration;

public class Dictionary {

    /** The locale of this dictionary. */
    protected Locale locale_ = null;
    
    /** The resource that has been loaded */
    protected String resource_ = null;

    /** The ResourceBundle which contains the raw key-value mappings. Loaded from a file named "resource_locale.properties*/
    private ResourceBundle raw_ = null;

    /**  Constructs the Dictionary class by first checking if a Locale has been set. If not the default locale is used, and a ResourceBundle is created. Finally a single important String, Language, is made available outside the class so a more read-able version of the Locale of this Dictionary is present.
     * @param locale The <strong>Locale</strong> used to load the desired dictionary resource bundle.
     */
    public Dictionary(String resource, String lang) {
	// Initialize.
	
	locale_ = new Locale(lang);
	resource_ = resource;
	if (locale_ == null) {
	    locale_ = Locale.getDefault();
	}
	try {
	    raw_ = ResourceBundle.getBundle(resource_, locale_);
	} catch (Exception e) {
	    System.err.println("Dictionary: couldn't locate a resource bundle. Error message: "+e.getMessage());
	}
    }

    public Enumeration getKeys() {
	if (raw_ != null) {
	    return raw_.getKeys();
	}
	return null;
    }
    /** Overloaded to call get with both a key and an empty argument array.
     * @param key A <strong>String</strong> which is mapped to a initial String within the ResourceBundle.
     * @return A <strong>String</strong> which has been referenced by the key String and that  contains no argument fields.
     */
    public String get(String key) {
	return get(key, null);
    }
    
    /** Used to retrieve a property value from the Locale specific ResourceBundle, based upon the key and arguments supplied. If the key cannot be found or if some other part of the call fails a default (English) error message is returned. <BR>
     * Here the get recieves a second argument which is an array of Strings used to populate argument fields, denoted {<I>n</I>}, within the value String returned.
     * @param key A <strong>String</strong> which is mapped to a initial String within the ResourceBundle.
     * @param args A <strong>String[]</strong> used to populate argument fields within the complete String.
     * @return A <strong>String</strong> which has been referenced by the key String and that either contains no argument fields, or has had the argument fields  populated  with argument Strings provided in the get call.
     */
    public String get(String key, String args[]) {
	if (raw_ == null) {
	    return null;
	}
	try {
	    String initial = raw_.getString(key);
	    // If the string contains arguments we have to insert them.
	    String complete = "";
	    // While we still have initial string left.
	    while(initial.length() > 0 && initial.indexOf('{') != -1 && initial.indexOf('}') != -1) {
		// Remove preamble
		int opening = initial.indexOf('{');
		int closing = initial.indexOf('}');
		complete = complete + initial.substring(0, opening);
		// Parse arg_num
		String arg_str = initial.substring(opening + 1, closing);
		int arg_num = Integer.parseInt(arg_str);
		if(closing + 1 < initial.length()) {
		    initial = initial.substring(closing + 1);
		}
		else {
		    initial = "";
		}
		// Insert argument
		if(args != null && 0 <= arg_num && arg_num < args.length) {
		    complete = complete + args[arg_num];
		}
	    }
	    return complete + initial;
	}
	catch (Exception e) {
	    System.err.println("Exception: " + e.toString());
	    e.printStackTrace();
	    return null;
	}
    }
}

